/*
 * Copyright (c) 1999, 2000, 2005, 2010 Greg Haerr <greg@censoft.com>
 * Copyright (c) 1991 David I. Bell
 *
 * Graphics server event routines for windows.
 *
 * Modifications:
 *  Date		Author			Description
 *	2003/09/24	Gabriele Brugnoni	In MwDeliverMouseEvent, if window is disabled,
 *                                      	(or wnd is child and parent disab), do nothing.
 *
 */
#include "windows.h"
#include "wintern.h"
#include "device.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

static LPFN_KEYBTRANSLATE mwPtrKeyboardTranslator = NULL;

typedef struct {
  CHAR16        UnicodeChar;
  WPARAM        WParam;
  BOOLEAN       ShiftState;
} UNICODE_VK_MAP;

STATIC UNICODE_VK_MAP mUnicodeToVkKeyTable[] = {
//
// Unicode  VK_CODE         SihftState     Index
//
  {0x00,    0x00,           FALSE},   ///< 0x00
  {0x01,    0x00,           FALSE},   ///< 0x01
  {0x02,    0x00,           FALSE},   ///< 0x02
  {0x03,    0x00,           FALSE},   ///< 0x03
  {0x04,    0x00,           FALSE},   ///< 0x04
  {0x05,    0x00,           FALSE},   ///< 0x05
  {0x06,    0x00,           FALSE},   ///< 0x06
  {0x07,    0x00,           FALSE},   ///< 0x07
  {0x08,    VK_BACK,        FALSE},   ///< 0x08
  {0x09,    VK_TAB,         FALSE},   ///< 0x09
  {0x0a,    0x00,           FALSE},   ///< 0x0a
  {0x0b,    0x00,           FALSE},   ///< 0x0b
  {0x0c,    0x00,           FALSE},   ///< 0x0c
  {0x0d,    VK_RETURN,      FALSE},   ///< 0x0d
  {0x0e,    0x00,           FALSE},   ///< 0x0e
  {0x0f,    0x00,           FALSE},   ///< 0x0f
  {0x10,    0x00,           FALSE},   ///< 0x10
  {0x11,    0x00,           FALSE},   ///< 0x11
  {0x12,    0x00,           FALSE},   ///< 0x12
  {0x13,    0x00,           FALSE},   ///< 0x13
  {0x14,    0x00,           FALSE},   ///< 0x14
  {0x15,    0x00,           FALSE},   ///< 0x15
  {0x16,    0x00,           FALSE},   ///< 0x16
  {0x17,    0x00,           FALSE},   ///< 0x17
  {0x18,    0x00,           FALSE},   ///< 0x18
  {0x19,    0x00,           FALSE},   ///< 0x19
  {0x1a,    0x00,           FALSE},   ///< 0x1a
  {0x1b,    0x00,           FALSE},   ///< 0x1b
  {0x1c,    0x00,           FALSE},   ///< 0x1c
  {0x1d,    0x00,           FALSE},   ///< 0x1d
  {0x1e,    0x00,           FALSE},   ///< 0x1e
  {0x1f,    0x00,           FALSE},   ///< 0x1f
  {L' ',    VK_SPACE,       FALSE},   ///< 0x20
  {L'!',    0x31,           TRUE },   ///< 0x21
  {L'"',    VK_OEM_7,       TRUE },   ///< 0x22
  {L'#',    0x33,           TRUE },   ///< 0x23
  {L'$',    0x34,           TRUE },   ///< 0x24
  {L'%',    0x35,           TRUE },   ///< 0x25
  {L'&',    0x37,           TRUE },   ///< 0x26
  {L'\'',   VK_OEM_7,       FALSE},    ///< 0x27
  {L'(',    0x39,           TRUE },   ///< 0x28
  {L')',    0x30,           TRUE },   ///< 0x29
  {L'*',    0x38,           TRUE },   ///< 0x2a
  {L'+',    VK_OEM_PLUS,    TRUE },   ///< 0x2b
  {L',',    VK_OEM_COMMA,   FALSE},   ///< 0x2c
  {L'-',    VK_OEM_MINUS,   FALSE},   ///< 0x2d
  {L'.',    VK_OEM_PERIOD,  FALSE},   ///< 0x2e
  {L'/',    VK_OEM_2,       FALSE},   ///< 0x2f
  {L'0',    0x30,           FALSE},   ///< 0x30
  {L'1',    0x31,           FALSE},   ///< 0x31
  {L'2',    0x32,           FALSE},   ///< 0x32
  {L'3',    0x33,           FALSE},   ///< 0x33
  {L'4',    0x34,           FALSE},   ///< 0x34
  {L'5',    0x35,           FALSE},   ///< 0x35
  {L'6',    0x36,           FALSE},   ///< 0x36
  {L'7',    0x37,           FALSE},   ///< 0x37
  {L'8',    0x38,           FALSE},   ///< 0x38
  {L'9',    0x39,           FALSE},   ///< 0x39
  {L':',    VK_OEM_1,       TRUE },   ///< 0x3a
  {L';',    VK_OEM_1,       FALSE},   ///< 0x3b
  {L'<',    VK_OEM_COMMA,   TRUE },   ///< 0x3c
  {L'=',    VK_OEM_PLUS,    FALSE},   ///< 0x3d
  {'>',     VK_OEM_PERIOD,  TRUE},    ///< 0x3e
  {'?',     VK_OEM_2,       TRUE},    ///< 0x3f
  {L'@',    0x32,           TRUE },   ///< 0x40
  {L'A',    0x41,           TRUE },   ///< 0x41
  {L'B',    0x42,           TRUE },   ///< 0x42
  {L'C',    0x43,           TRUE },   ///< 0x43
  {L'D',    0x44,           TRUE },   ///< 0x44
  {L'E',    0x45,           TRUE },   ///< 0x45
  {L'F',    0x46,           TRUE },   ///< 0x46
  {L'G',    0x47,           TRUE },   ///< 0x47
  {L'H',    0x48,           TRUE },   ///< 0x48
  {L'I',    0x49,           TRUE },   ///< 0x49
  {L'J',    0x4a,           TRUE },   ///< 0x4a
  {L'K',    0x4b,           TRUE },   ///< 0x4b
  {L'L',    0x4c,           TRUE },   ///< 0x4c
  {L'M',    0x4d,           TRUE },   ///< 0x4d
  {L'N',    0x4e,           TRUE },   ///< 0x4e
  {L'O',    0x4f,           TRUE },   ///< 0x4f
  {L'P',    0x50,           TRUE },   ///< 0x50
  {L'Q',    0x51,           TRUE },   ///< 0x51
  {L'R',    0x52,           TRUE },   ///< 0x52
  {L'S',    0x53,           TRUE },   ///< 0x53
  {L'T',    0x54,           TRUE },   ///< 0x54
  {L'U',    0x55,           TRUE },   ///< 0x55
  {L'V',    0x56,           TRUE },   ///< 0x56
  {L'W',    0x57,           TRUE },   ///< 0x57
  {L'X',    0x58,           TRUE },   ///< 0x58
  {L'Y',    0x59,           TRUE },   ///< 0x59
  {L'Z',    0x5a,           TRUE },   ///< 0x5a
  {L'[',    VK_OEM_4,       FALSE},   ///< 0x5b
  {L'\\',   VK_OEM_5,       FALSE},   ///< 0x5c
  {L']',    VK_OEM_6,       FALSE},   ///< 0x5d
  {L'^',    0x36,           TRUE },   ///< 0x5e
  {L'_',    VK_OEM_MINUS,   TRUE },   ///< 0x5f
  {L'`',    VK_OEM_3,       FALSE},   ///< 0x60
  {L'a',    0x41,           FALSE},   ///< 0x61
  {L'b',    0x42,           FALSE},   ///< 0x62
  {L'c',    0x43,           FALSE},   ///< 0x63
  {L'd',    0x44,           FALSE},   ///< 0x64
  {L'e',    0x45,           FALSE},   ///< 0x65
  {L'f',    0x46,           FALSE},   ///< 0x66
  {L'g',    0x47,           FALSE},   ///< 0x67
  {L'h',    0x48,           FALSE},   ///< 0x68
  {L'i',    0x49,           FALSE},   ///< 0x69
  {L'j',    0x4a,           FALSE},   ///< 0x6a
  {L'k',    0x4b,           FALSE},   ///< 0x6b
  {L'l',    0x4c,           FALSE},   ///< 0x6c
  {L'm',    0x4d,           FALSE},   ///< 0x6d
  {L'n',    0x4e,           FALSE},   ///< 0x6e
  {L'o',    0x4f,           FALSE},   ///< 0x6f
  {L'p',    0x50,           FALSE},   ///< 0x70
  {L'q',    0x51,           FALSE},   ///< 0x71
  {L'r',    0x52,           FALSE},   ///< 0x72
  {L's',    0x53,           FALSE},   ///< 0x73
  {L't',    0x54,           FALSE},   ///< 0x74
  {L'u',    0x55,           FALSE},   ///< 0x75
  {L'v',    0x56,           FALSE},   ///< 0x76
  {L'w',    0x57,           FALSE},   ///< 0x77
  {L'x',    0x58,           FALSE},   ///< 0x78
  {L'y',    0x59,           FALSE},   ///< 0x79
  {L'z',    0x5a,           FALSE},   ///< 0x7a
  {L'{',    VK_OEM_4,       TRUE },   ///< 0x7b
  {L'|',    VK_OEM_5,       TRUE },   ///< 0x7c
  {L'}',    VK_OEM_6,       TRUE },   ///< 0x7d
  {L'~',    VK_OEM_3,       TRUE },   ///< 0x7e
  {0x7f,    0x00,           FALSE}    ///< 0x7f
};

/*
 * Update mouse status and issue events on it if necessary.
 * This function doesn't block, but is normally only called when
 * there is known to be some data waiting to be read from the mouse.
 */
BOOL
MwCheckMouseEvent(void)
{
	MWCOORD		rootx;		/* latest mouse x position */
	MWCOORD		rooty;		/* latest mouse y position */
	MWCOORD		rootz;    /* latest mouse x position */
	int		newbuttons;	/* latest buttons */
	int		mousestatus;	/* latest mouse status */

	/* Read the latest mouse status: */
	mousestatus = GdReadMouse(&rootx, &rooty, &rootz,  &newbuttons);
	if(mousestatus < 0) {
		/*MwError(GR_ERROR_MOUSE_ERROR, 0);*/
		return FALSE;
	} else if(mousestatus) {	/* Deliver events as appropriate: */
    MwHandleMouseStatus(rootx, rooty, rootz, newbuttons);
		return TRUE;
	}
	return FALSE;
}

/*
 * Update keyboard status and issue events on it if necessary.
 * This function doesn't block, but is normally only called when
 * there is known to be some data waiting to be read from the keyboard.
 */
BOOL
MwCheckKeyboardEvent(void)
{
	EFI_KEY_DATA      KeyData;
	MWKEYMOD 	      modifiers;	/* latest modifiers */
	MWSCANCODE	      scancode;

	int	 	keystatus;	/* latest keyboard status */

	/* Read the latest keyboard status: */
  keystatus = GdReadKeyboard(&KeyData, &modifiers, &scancode);
	if(keystatus < 0) {
		if(keystatus == -2)	/* special case for ESC pressed*/
			MwTerminate();
		/*MwError(GR_ERROR_KEYBOARD_ERROR, 0);*/
		return FALSE;
	} else if(keystatus) {		/* Deliver events as appropriate: */
#if 0
		switch (mwkey) {
		case MWKEY_QUIT:
#if DEBUG
			if (modifiers & MWKMOD_CTRL)
				GdCaptureScreen(NULL, "screen.bmp");
			else
#endif
				MwTerminate();
			break;
		case MWKEY_REDRAW:
			MwRedrawScreen();
			break;
		case MWKEY_PRINT:
#if DEBUG
			if (keystatus == 1)
				GdCaptureScreen(NULL, "screen.bmp");
#endif
			break;
		}
#endif
		MwDeliverKeyboardEvent(&KeyData, modifiers, scancode, keystatus == 1? TRUE: FALSE);
		return TRUE;
	}
	return FALSE;
}

/*
 * Handle all mouse events.  These are mouse enter, mouse exit, mouse
 * motion, mouse position, button down, and button up.  This also moves
 * the cursor to the new mouse position and changes it shape if needed.
 */
void
MwHandleMouseStatus(MWCOORD newx, MWCOORD newy, MWCOORD newz, int newbuttons)
{
	int		changebuttons;	/* buttons that have changed */
	MWKEYMOD	modifiers;	/* latest modifiers */
	static int curbuttons;
	modifiers = 0;
	GdGetModifierInfo(NULL, &modifiers); /* Read kbd modifiers */

	/*
	 * First, if the mouse has moved, then position the cursor to the
	 * new location, which will send mouse enter, mouse exit, focus in,
	 * and focus out events if needed.  Check here for mouse motion and
	 * mouse position events.
	 */
	if (newx != cursorx || newy != cursory) {
		MwMoveCursor(newx, newy);
		MwDeliverMouseEvent(newbuttons, 0, modifiers);
	}

	if (newz != cursorz) {
		cursorz = newz;
		MwDeliverMouseEvent(newbuttons, MWBUTTON_WHEEL, modifiers);
		cursorz = 0;
	}


	/*
	 * Next, generate a button up event if any buttons have been released.
	 */
	changebuttons = (curbuttons & ~newbuttons);
	if (changebuttons)
		MwDeliverMouseEvent(newbuttons, changebuttons, modifiers);

	/*
	 * Finally, generate a button down event if any buttons have been
	 * pressed.
	 */
	changebuttons = (~curbuttons & newbuttons);
	if (changebuttons)
		MwDeliverMouseEvent(newbuttons, changebuttons, modifiers);

	curbuttons = newbuttons;
}

/*
 * Translate and deliver hardware mouse message to proper window.
 */
void
MwTranslateMouseMessage(HWND hwnd,UINT msg,int hittest,int buttons)
{
	POINT		pt;
	UINT64		tick;
	static UINT	lastmsg = 0;
	static HWND	lasthwnd;
	static UINT64	lasttick;
	static int	lastx, lasty;

	/* determine double click eligibility*/
	if(msg == WM_LBUTTONDOWN || msg == WM_RBUTTONDOWN) {
		tick = GetTickCount();
		if((hwnd->pClass->style & CS_DBLCLKS) &&
		    msg == lastmsg && hwnd == lasthwnd &&
		    ((tick - lasttick) < DBLCLICKSPEED) &&
		    abs(cursorx-lastx) < mwSYSMETRICS_CXDOUBLECLK &&
		    abs(cursory-lasty) < mwSYSMETRICS_CYDOUBLECLK)
			msg += (WM_LBUTTONDBLCLK - WM_LBUTTONDOWN);
		lastmsg = msg;
		lasthwnd = hwnd;
		lasttick = tick;
		lastx = cursorx;
		lasty = cursory;
	}

	/*
	 * We always send nc mouse message
	 * unlike Windows, for HTCLIENT default processing
	 */
	PostMessage(hwnd, msg + (WM_NCMOUSEMOVE-WM_MOUSEMOVE), hittest,
		MAKELONG(cursorx, cursory));

	/* then possibly send user mouse message*/
	if(hittest == HTCLIENT) {
		pt.x = cursorx;
		pt.y = cursory;
		ScreenToClient(hwnd, &pt);
		if (msg == WM_MOUSEWHEEL) {
			PostMessage (hwnd, msg, MAKELONG(buttons, cursorz), MAKELONG(pt.x, pt.y));
		} else {
			PostMessage(hwnd, msg, buttons, MAKELONG(pt.x, pt.y));
		}
	}
}

/*
 * Deliver a mouse button or motion event.
 */
int mwCurrentButtons;

HWND
MwFindChildTestWnd (
  HWND                          Wnd,
  INT32                         *Hittest
  )
{
  INT32                         Result;
  HWND                          Child;
  POINT                         Point;

  Child = Wnd->children;
  Point.x = cursorx;
  Point.y = cursory;
  for (Child = Wnd->children; Child != NULL; Child = Child->siblings) {
    if (Child->unmapcount || !PtInRect (&(Child->winrect), Point)) {
      continue;
    }

    if (Child->children != NULL) {
      Wnd = MwFindChildTestWnd (Child, Hittest);
      if (Wnd != NULL) {
        return Wnd;
      }
    }

    Result = (UINT32) SendMessage(Child, WM_NCHITTEST, 0, MAKELONG(cursorx,cursory));
    if (Result != HTTRANSPARENT) {
      *Hittest = Result;
      return Child;
    }
  }

  return NULL;
}

void
MwDeliverMouseEvent(int buttons, int changebuttons, MWKEYMOD modifiers)
{
	HWND	hwnd, top;
	int	hittest;
	UINT	msg;
  HWND                          Parent;

	mwCurrentButtons = buttons;

	hwnd = GetCapture();
	if(!hwnd)
		hwnd = mousewp;

	/* if wnd is disabled, or window is child and parent disabled, do nothing*/
	if( hwnd == NULL || !IsWindowEnabled(hwnd) )
		return;

	top = MwGetTopWindow(hwnd);
	if( top != NULL && !IsWindowEnabled(top) )
		return;

  hittest = (UINT32) SendMessage(hwnd, WM_NCHITTEST, 0, MAKELONG(cursorx,cursory));
  Parent = hwnd->parent;
  while (hittest == HTTRANSPARENT) {
    hwnd = MwFindChildTestWnd (Parent, &hittest);
    Parent = Parent->parent;
  }
  ASSERT (hwnd);

	if(!changebuttons)
		MwTranslateMouseMessage(hwnd, WM_MOUSEMOVE, hittest, buttons);

	if (changebuttons & MWBUTTON_WHEEL) {
		msg = WM_MOUSEWHEEL;
		MwTranslateMouseMessage(hwnd, msg, hittest, buttons);
	}

	if(changebuttons & MWBUTTON_L) {
		msg = (buttons&MWBUTTON_L)? WM_LBUTTONDOWN: WM_LBUTTONUP;
		MwTranslateMouseMessage(hwnd, msg, hittest, buttons);
	}

	if(changebuttons & MWBUTTON_M) {
		msg = (buttons&MWBUTTON_M)? WM_MBUTTONDOWN: WM_MBUTTONUP;
		MwTranslateMouseMessage(hwnd, msg, hittest, buttons);
	}

	if(changebuttons & MWBUTTON_R) {
		msg = (buttons&MWBUTTON_R)? WM_RBUTTONDOWN: WM_RBUTTONUP;
		MwTranslateMouseMessage(hwnd, msg, hittest, buttons);
	}
}

/*
 *  Keyboard filter function
 */
void WINAPI
MwSetKeyboardTranslator(LPFN_KEYBTRANSLATE pFn)
{
	mwPtrKeyboardTranslator = pFn;
}

BOOLEAN
MwCovertUniCodeToVkKey (
  IN   EFI_KEY_DATA       *KeyData,
  OUT  WPARAM             *WParam,
  OUT  LPARAM             *LParam
  )
{
  CHAR16             UnicodeChar;
  UINT32             KeyShiftState;

  ASSERT (KeyData != NULL);
  ASSERT (WParam  != NULL);
  ASSERT (LParam  != NULL);

  UnicodeChar = KeyData->Key.UnicodeChar;
  if (UnicodeChar == 0 || UnicodeChar >= 0x7F) {
    return FALSE;
  }

  *WParam = mUnicodeToVkKeyTable[UnicodeChar].WParam;
  if (mUnicodeToVkKeyTable[UnicodeChar].ShiftState) {
    //
    // using a reserved bit (bit 25) in lparam to indicate the shift key is pressed
    //
    *LParam |= (1 << 25);
  }

  KeyShiftState = KeyData->KeyState.KeyShiftState;
  if ((KeyShiftState & EFI_SHIFT_STATE_VALID) != 0 &&
      (KeyShiftState & (EFI_RIGHT_ALT_PRESSED | EFI_LEFT_ALT_PRESSED)) != 0) {
    //
    // Enable context code bit to indicate ALT key is pressed.
    //
    *LParam |= (1 << 29);
  }

  if ((KeyShiftState & EFI_SHIFT_STATE_VALID) != 0 &&
      (KeyShiftState & (EFI_RIGHT_CONTROL_PRESSED | EFI_LEFT_CONTROL_PRESSED)) != 0) {
    //
    // Using a reserved bit (bit 26) in lparam to indicate control key is pressed.
    //
    *LParam |= (1 << 26);
  }
  return TRUE;


}

BOOLEAN
MwCovertScanCodeToVkKey (
  IN   EFI_INPUT_KEY      *Key,
  OUT  WPARAM             *WParam
  )
{
  BOOLEAN      Convert;

  Convert = FALSE;
  switch (Key->ScanCode) {
    case SCAN_HOME:       *WParam = VK_HOME;      Convert = TRUE;  break;
    case SCAN_END:        *WParam = VK_END;       Convert = TRUE;  break;
    case SCAN_LEFT:       *WParam = VK_LEFT;      Convert = TRUE;  break;
    case SCAN_RIGHT:      *WParam = VK_RIGHT;     Convert = TRUE;  break;
    case SCAN_UP:         *WParam = VK_UP;        Convert = TRUE;  break;
    case SCAN_DOWN:       *WParam = VK_DOWN;      Convert = TRUE;  break;
    case SCAN_DELETE:     *WParam = VK_DELETE;    Convert = TRUE;  break;
    case SCAN_INSERT:     *WParam = VK_INSERT;    Convert = TRUE;  break;
    case SCAN_PAGE_UP:    *WParam = VK_PRIOR;     Convert = TRUE;  break;
    case SCAN_PAGE_DOWN:  *WParam = VK_NEXT;      Convert = TRUE;  break;
    case SCAN_ESC:        *WParam = VK_ESCAPE;    Convert = TRUE;  break;

    case SCAN_F1:         *WParam = VK_F1;        Convert = TRUE;  break;
    case SCAN_F2:         *WParam = VK_F2;        Convert = TRUE;  break;
    case SCAN_F3:         *WParam = VK_F3;        Convert = TRUE;  break;
    case SCAN_F4:         *WParam = VK_F4;        Convert = TRUE;  break;
    case SCAN_F5:         *WParam = VK_F5;        Convert = TRUE;  break;
    case SCAN_F6:         *WParam = VK_F6;        Convert = TRUE;  break;
    case SCAN_F7:         *WParam = VK_F7;        Convert = TRUE;  break;
    case SCAN_F8:         *WParam = VK_F8;        Convert = TRUE;  break;
    case SCAN_F9:         *WParam = VK_F9;        Convert = TRUE;  break;
    case SCAN_F10:        *WParam = VK_F10;       Convert = TRUE;  break;
    case SCAN_F11:        *WParam = VK_F11;       Convert = TRUE;  break;
    case SCAN_F12:        *WParam = VK_F12;       Convert = TRUE;  break;

    case SCAN_F13:        *WParam = VK_F13;       Convert = TRUE;  break;
    case SCAN_F14:        *WParam = VK_F14;       Convert = TRUE;  break;
    case SCAN_F15:        *WParam = VK_F15;       Convert = TRUE;  break;
    case SCAN_F16:        *WParam = VK_F16;       Convert = TRUE;  break;
    case SCAN_F17:        *WParam = VK_F17;       Convert = TRUE;  break;
    case SCAN_F18:        *WParam = VK_F18;       Convert = TRUE;  break;
    case SCAN_F19:        *WParam = VK_F19;       Convert = TRUE;  break;
    case SCAN_F20:        *WParam = VK_F20;       Convert = TRUE;  break;
    case SCAN_F21:        *WParam = VK_F21;       Convert = TRUE;  break;
    case SCAN_F22:        *WParam = VK_F22;       Convert = TRUE;  break;
    case SCAN_F23:        *WParam = VK_F23;       Convert = TRUE;  break;
    case SCAN_F24:        *WParam = VK_F24;       Convert = TRUE;  break;
    case SCAN_PAUSE:      *WParam = VK_PAUSE;     Convert = TRUE;  break;
  }
  return Convert;
}
/*
 * Deliver a keyboard event.
 */
void
MwDeliverKeyboardEvent(EFI_KEY_DATA   *KeyData, MWKEYMOD modifiers, MWSCANCODE scancode,
	BOOL pressed)
{
  EFI_INPUT_KEY      *Key;
	WPARAM             VK_Code = (WPARAM)(-1);    /* set to default if no VK found. */
  LPARAM             lParam = 0L;		            /* used to specify control keys */

#if 0
	/* Keysyms from 1-255 are mapped to ASCII*/
	if (keyvalue < 1 || keyvalue > 255)
	  switch(keyvalue) {

	/* arrows + home/end pad*/
	case MWKEY_LEFT:
		VK_Code = VK_LEFT;
		break;
	case MWKEY_RIGHT:
		VK_Code = VK_RIGHT;
		break;
	case MWKEY_UP:
		VK_Code = VK_UP;
		break;
	case MWKEY_DOWN:
		VK_Code =  VK_DOWN;
		break;
	case MWKEY_INSERT:
		VK_Code = VK_INSERT;
		break;
	case MWKEY_DELETE:
		VK_Code = VK_DELETE;
		break;
	case MWKEY_HOME:
		VK_Code = VK_HOME;
		break;
	case MWKEY_END:
		VK_Code = VK_END;
		break;
	case MWKEY_PAGEUP:
		VK_Code = VK_PRIOR;
		break;
	case MWKEY_PAGEDOWN:
		VK_Code = VK_NEXT;
		break;

	/* Numeric keypad*/
	case MWKEY_KP0:
		VK_Code = VK_NUMPAD0;
		break;
	case MWKEY_KP1:
		VK_Code = VK_NUMPAD1;
		break;
	case MWKEY_KP2:
		VK_Code = VK_NUMPAD2;
		break;
	case MWKEY_KP3:
		VK_Code = VK_NUMPAD3;
		break;
	case MWKEY_KP4:
		VK_Code = VK_NUMPAD4;
		break;
	case MWKEY_KP5:
		VK_Code = VK_NUMPAD5;
		break;
	case MWKEY_KP6:
		VK_Code = VK_NUMPAD6;
		break;
	case MWKEY_KP7:
		VK_Code = VK_NUMPAD7;
		break;
	case MWKEY_KP8:
		VK_Code = VK_NUMPAD8;
		break;
	case MWKEY_KP9:
		VK_Code = VK_NUMPAD9;
		break;
	case MWKEY_KP_PERIOD:
		VK_Code = VK_DECIMAL;
		break;
	case MWKEY_KP_DIVIDE:
		VK_Code = VK_DIVIDE;
		break;
	case MWKEY_KP_MULTIPLY:
		VK_Code = VK_MULTIPLY;
		break;
	case MWKEY_KP_MINUS:
		VK_Code = VK_SUBTRACT;
		break;
	case MWKEY_KP_PLUS:
		VK_Code = VK_ADD;
		break;
	case MWKEY_KP_ENTER:
		VK_Code = VK_RETURN;
		break;

	/* Function keys */
	case MWKEY_F1:
		VK_Code = VK_F1;
		break;
	case MWKEY_F2:
		VK_Code = VK_F2;
		break;
	case MWKEY_F3:
		VK_Code = VK_F3;
		break;
	case MWKEY_F4:
		VK_Code = VK_F4;
		break;
	case MWKEY_F5:
		VK_Code = VK_F5;
		break;
	case MWKEY_F6:
		VK_Code = VK_F6;
		break;
	case MWKEY_F7:
		VK_Code = VK_F7;
		break;
	case MWKEY_F8:
		VK_Code = VK_F8;
		break;
	case MWKEY_F9:
		VK_Code = VK_F9;
		break;
	case MWKEY_F10:
		VK_Code = VK_F10;
		break;
	case MWKEY_F11:
		VK_Code = VK_F11;
		break;
	case MWKEY_F12:
		VK_Code = VK_F12;
		break;

	/* Key state modifier keys*/
	case MWKEY_NUMLOCK:
		VK_Code = VK_NUMLOCK;
		break;
	case MWKEY_CAPSLOCK:
		VK_Code = VK_CAPITAL;
		break;
	case MWKEY_SCROLLOCK:
		VK_Code = VK_CAPITAL;
		break;
	case MWKEY_LSHIFT:
		VK_Code = VK_LSHIFT;
		break;
	case MWKEY_RSHIFT:
		VK_Code = VK_RSHIFT;
		break;
	case MWKEY_LCTRL:
		VK_Code = VK_LCONTROL;
		break;
	case MWKEY_RCTRL:
		VK_Code = VK_RCONTROL;
		break;
	case MWKEY_LALT:
		VK_Code = VK_MENU;
		break;
	case MWKEY_RALT:
		VK_Code = VK_MENU;
		break;
	case MWKEY_LMETA:
		VK_Code = VK_F1;	// Smartphone 2003 compliance
		break;
	case MWKEY_RMETA:
		VK_Code = VK_F2;	// Smartphone 2003 compliance
		break;
	case MWKEY_ACCEPT:
		VK_Code = VK_RETURN;// Smartphone 2003 compliance
		break;

	/* Misc function keys*/
	case MWKEY_PRINT:
		VK_Code = VK_PRINT;
		break;
	case MWKEY_PAUSE:
		VK_Code = VK_PAUSE;
		break;
	case MWKEY_MENU:
		VK_Code = VK_LMENU;	/* virtual key*/
		break;

	/* questionable mappings or no mappings...*/
	case MWKEY_KP_EQUALS:
		VK_Code = '=';	/* FIXME*/
		break;

	/* map all non-handled MWKEY values to VK_NONAME*/
#if 0
	case MWKEY_UNKNOWN:
	case MWKEY_SYSREQ:
	case MWKEY_BREAK
	case MWKEY_QUIT:
	case MWKEY_REDRAW:
	case MWKEY_ALTGR:
	/* Handheld function keys*/
	case MWKEY_RECORD:
	case MWKEY_PLAY:
	case MWKEY_CONTRAST:
	case MWKEY_BRIGHTNESS:
	case MWKEY_SELECTUP:
	case MWKEY_SELECTDOWN:
	case MWKEY_CANCEL:
	case MWKEY_APP1:
	case MWKEY_APP2:
	case MWKEY_LAST:
#endif
	default:
		VK_Code = VK_NONAME;
		break;
	}
#else

  Key           = &KeyData->Key;
  if (Key->ScanCode != SCAN_NULL) {
    VK_Code = VK_NONAME;
    MwCovertScanCodeToVkKey (Key, &VK_Code);
  }
#endif
	/* if no VK defined, set to keyvalue default*/
  if( VK_Code == -1 ) {
    MwCovertUniCodeToVkKey (KeyData, &VK_Code, &lParam);
  } else {
    lParam |= (1 << 24);	/* set control bit in lParam*/
  }

  if (mwPtrKeyboardTranslator) {
    mwPtrKeyboardTranslator(&VK_Code, &lParam, &pressed);
  }

  if (!MwDeliverHotkey (VK_Code, pressed)) {
    if (pressed) {
      PostMessage(focuswp, WM_KEYDOWN, VK_Code, lParam);
    } else {
      PostMessage(focuswp, WM_KEYUP, VK_Code, lParam);
    }
  }
}

/*
 * Deliver a window expose event.
 * Most of the work is in calculating the update region
 * for better redraw look and feel, and then queuing a
 * WM_PAINT message to the window.
 */
void
MwDeliverExposureEvent(HWND wp, MWCOORD x, MWCOORD y, MWCOORD width,
	MWCOORD height)
{
	if (wp->unmapcount)
		return;

	MwUnionUpdateRegion(wp, x, y, width, height, TRUE);
	PostMessage(wp, WM_PAINT, 0, 0L);
}

/*
 * Combine the passed rectangle with the update region for the given window.
 * Coordinates are passed relative to window.
 * If bUnion is TRUE, union the rectangle, otherwise subtract it.
 */
void
MwUnionUpdateRegion(HWND wp, MWCOORD x, MWCOORD y, MWCOORD width,
	MWCOORD height, BOOL bUnion)
{
#if UPDATEREGIONS
	RECT rc;

	if (wp->unmapcount)
		return;

	/* convert window relative coords to screen coords*/
	rc.left = x + wp->winrect.left;
	rc.top = y + wp->winrect.top;
	rc.right = rc.left + width;
	rc.bottom = rc.top + height;
	wp->nEraseBkGnd++;

	if(bUnion)
		GdUnionRectWithRegion(&rc, wp->update);
	else
		GdSubtractRectFromRegion(&rc, wp->update);
#endif
}

/*
 * Move the cursor to the specified absolute screen coordinates.
 * The coordinates are that of the defined hot spot of the cursor.
 * The cursor's appearance is changed to that defined for the window
 * in which the cursor is moved to.  In addition, the window the
 * cursor is in is recalculated.
 */
void MwMoveCursor(MWCOORD x, MWCOORD y)
{
	/*
	 * Move the cursor only if necessary, offsetting it to
	 * place the hot spot at the specified coordinates.
	 */
	if (x != cursorx || y != cursory) {
		if(curcursor)
			GdMoveCursor(x - curcursor->cursor.hotx,
				y - curcursor->cursor.hoty);
		cursorx = x;
		cursory = y;
	}

	/*
	 * Now check to see which window the mouse is in and whether or
	 * not the cursor shape should be changed.
	 */
	MwCheckMouseWindow();
	MwCheckCursor();
}

/*
 * Check to see if the cursor shape is the correct shape for its current
 * location.  If not, its shape is changed.
 */
void MwCheckCursor(void)
{
	HWND 		wp;		/* window cursor is in */
	HCURSOR		cp;		/* cursor definition */

	/*
	 * Get the cursor at its current position, and if it is not the
	 * currently defined one, then set the new cursor.  However,
	 * if the window is currently captured, then leave it alone.
	 */
	wp = capturewp;
	if (wp == NULL)
		wp = mousewp;

	cp = wp->cursor;
	if (cp == curcursor)
		return;

	/*
	 * It needs redefining, so do it.
	 */
	curcursor = cp;
	GdMoveCursor(cursorx - cp->cursor.hotx, cursory - cp->cursor.hoty);
	GdSetCursor(&cp->cursor);
}

/*
 * Find the window which is currently visible for the specified coordinates.
 * This just walks down the window tree looking for the deepest mapped
 * window which contains the specified point.  If the coordinates are
 * off the screen, the root window is returned.
 */
HWND
MwFindVisibleWindow(MWCOORD x, MWCOORD y)
{
	HWND	wp;		/* current window */
	HWND	retwp;		/* returned window */

	wp = rootwp;
	retwp = wp;
	while (wp) {
		if (!wp->unmapcount &&
		    wp->winrect.left <= x && wp->winrect.top <= y &&
		    wp->winrect.right > x && wp->winrect.bottom > y) {
			retwp = wp;
			wp = wp->children;
			continue;
		}
		wp = wp->siblings;
	}
	return retwp;
}

/*
 * Check to see if the window the mouse is currently in has changed.
 */
void MwCheckMouseWindow(void)
{
	HWND	wp;

	/* Don't change if window drag or capture in progress*/
	wp = dragwp;
	if(!wp)
		wp = capturewp;
	if(!wp)
		wp = MwFindVisibleWindow(cursorx, cursory);

	if (wp != NULL) {
		if (mousewp == NULL) {
			PostMessage (wp, WM_MOUSEENTER, (WPARAM)NULL, (LPARAM)MAKELONG(cursorx, cursory));
			mousewp = wp;
		} else if (mousewp != wp) {
			PostMessage (mousewp, WM_MOUSELEAVE, (WPARAM)NULL, (LPARAM)MAKELONG(cursorx, cursory));
			PostMessage (wp,      WM_MOUSEENTER, (WPARAM)NULL, (LPARAM)MAKELONG(cursorx, cursory));
		}
	}

	mousewp = wp;
}

/*
 *  Copy dstsiz bytes, including nul, from src to dst.
 *  Return # bytes, excluding nul, copied.
 */
int
strzcpy(char *dst,const char *src,int dstsiz)
{
	int	cc = dstsiz;

	/* return 0 on NULL src*/
	if(!src)
		cc = dstsiz = 1;

	while(--dstsiz > 0) {
		if((*dst++ = *src++) == '\0')
			return cc - dstsiz - 1;
	}
	*dst = 0;
	return cc - dstsiz - 1;
}

